from datetime import datetime
import os
from pathlib import Path
import sys

from OpenSSL.crypto import FILETYPE_PEM, FILETYPE_TEXT, load_certificate, \
    dump_certificate, X509, X509Store, X509StoreContext, \
    X509StoreContextError
from OpenSSL import SSL

from cryptography.hazmat.primitives.serialization import pkcs12
from cryptography.hazmat.primitives import serialization

from cryptography.hazmat.backends import default_backend
from cryptography import x509 as cryptography_x509


import pem
import magic
import socket

from .getremote import get_remote_certs
from .exceptions import CertException, InvalidCertificate
from .verifychain import verify_chain
from .printcert import print_cert, print_full_cert, print_dnames, print_names



def is_local(cert, net):
    """ guesses is cert is local file or not """
    if net:
        return False

    if os.path.exists(cert) and os.path.isfile(cert) or cert == '-':
        return True
    return False

def crypto_cert_to_pyopenssl(cert: cryptography_x509.Certificate) -> X509:
    pem_bytes = cert.public_bytes(encoding=serialization.Encoding.PEM)
    return load_certificate(FILETYPE_PEM, pem_bytes)

def get_local_certs(CERT, insecure=False, password=None):
    if CERT == '-':
        rawcert = sys.stdin.buffer.read()
    else:        
        rawcert = Path(CERT).read_bytes()

    mime = _detect_mimetype(rawcert)

    if mime.startswith("text/") or "BEGIN CERTIFICATE" in str(rawcert):
        # default simple PEM part
        return [ load_certificate(FILETYPE_PEM, str(_c)) for _c in pem.parse(rawcert) if _c.__class__.__name__ == 'Certificate' ]

    elif mime in ("application/x-pkcs12", "application/octet-stream"):
        # PKCS#12 part
        key, cert, addl = pkcs12.load_key_and_certificates(
            rawcert, password.encode() if password else None, default_backend()
        )
        certs = []
        if cert:
            certs.append(crypto_cert_to_pyopenssl(cert))
        if addl:
            certs.extend(crypto_cert_to_pyopenssl(c) for c in addl)
        return certs


def _detect_mimetype(file_contents: bytes) -> str:
    if hasattr(magic, "from_buffer"):
        # https://pypi.org/project/python-magic/
        return magic.from_buffer(file_contents, mime=True)
    # https://pypi.org/project/file-magic/
    _file_magic = magic.detect_from_content(file_contents) # pragma: no cover
    return _file_magic.mime_type # pragma: no cover


def get_days_left(crt):
    nafter = datetime.strptime(crt.get_notAfter().decode(), '%Y%m%d%H%M%SZ')
    left = (nafter - datetime.now()).days
    return left



def process_cert(CERT, name=None, insecure=False, warn=None, starttls='auto', output='brief',
                 force_network=False, trusted_ca=None, limit=None, chain=False, password=None, ipv4=False):

    retcode = 0
    hostname = None
    out = output.lower()
    sock_host = None
    path = None
    verified = None

    try:
        if is_local(CERT, net=force_network):
            cert_chain = get_local_certs(CERT, password=password)

            path = CERT
            if not cert_chain:
                raise InvalidCertificate('Can not load certificate from file '+ CERT + '\nUse option --net if you want to check host ' + CERT)

        else:
            hostname = name or CERT.split(':')[0]
            sock_host, cert_chain = get_remote_certs(remote=CERT, name=name, insecure=insecure, 
                                                starttls=starttls, trusted_ca=trusted_ca, limit=limit, ipv4=ipv4)
    except SSL.Error as e:
        print("{CERT} Certificate verification error (use -i): {exception}".format(CERT=CERT, exception=e),
            file=sys.stderr)
        return 1
    except socket.timeout as e:
        print("Timeout with {}".format(CERT), file=sys.stderr)
        return 1
    except socket.gaierror as e:
        print("Error with {}: {}".format(CERT, e), file=sys.stderr)
        return 1
    except ConnectionRefusedError as e:
        print(f"Connection refused to {CERT}", file=sys.stderr)
        return 1



    if not insecure:
        try:
            verify_chain(cert_chain, hostname, trusted_ca=trusted_ca)
            verified = True
        except X509StoreContextError as e:
            print("Verification error (use -i):", e.args[0])
            return 1
        except InvalidCertificate as e:
            print("Verification error (use -i):", e)
            return 1


    if not chain:
        cert_chain = cert_chain[0:1]


    # set retcode
    left = get_days_left(cert_chain[0])
    if warn is not None and left<warn:
        print("{CERT} expires in {left} days".format(CERT=CERT, left=left),
            file=sys.stderr)
        retcode = 2


    # output
    if out == 'raw' or out == 'pem':
        for _c in cert_chain:
            r = dump_certificate(FILETYPE_PEM, _c)
            print(r.decode(), end='')
        return 0
    elif out in ['brief', 'ext', 'extended']:
        for i,_c in enumerate(cert_chain):
            if i>0:
                print()
            print_cert(crt = _c, fmt=out, addr=sock_host, path=path, verified=verified)

    elif out == 'full':
        for i,_c in enumerate(cert_chain):
            if i>0:
                print()
            print_full_cert(_c)
    
    elif out == 'names':
        for i,_c in enumerate(cert_chain):
            if i>0:
                print()
            print_names(_c)

    elif out == 'dnames':
        for i,_c in enumerate(cert_chain):
            if i>0:
                print()
            print_dnames(_c)

    elif out == 'no':
        pass

    else:
        print("unknown output (-o) format", file=sys.stderr)

    return retcode




import socket
import certifi
import time
from OpenSSL import SSL
from collections import namedtuple

from .exceptions import ServerError, InvalidAddress

phrase = namedtuple('Phrase', 'say wait expect')

# not covering conversation because it's hard to find server which would produce protocol errors



def recv_until_newline(sock: socket.socket, timeout: float = 5.0) -> bytes: # pragma: no cover
    """Read from socket until '\n' seen or timeout expires."""
    sock.setblocking(False)
    data = bytearray()
    deadline = time.monotonic() + timeout

    while time.monotonic() < deadline:
        try:
            chunk = sock.recv(4096)
            if not chunk:  # connection closed
                break
            data.extend(chunk)
            if b'\n' in chunk:
                break
        except BlockingIOError:
            time.sleep(0.01)
            continue
    return bytes(data)

def recv_smtp(sock: socket.socket, timeout: float = 5.0) -> bytes: # pragma: no cover
    """
    Read full SMTP reply (single or multi-line) until final line received or timeout.
    RFC 5321: lines start with 3 digits + ('-' for continuation or ' ' for end).
    """
    sock.setblocking(False)
    data = bytearray()
    lines = []
    deadline = time.monotonic() + timeout
    code = None

    while time.monotonic() < deadline:
        try:
            chunk = sock.recv(4096)
            if not chunk:  # connection closed
                break
            data.extend(chunk)
            while b'\n' in data:
                line, _, rest = data.partition(b'\n')
                data = bytearray(rest)
                line = line.rstrip(b'\r')
                lines.append(line)

                # parse reply code
                if len(line) >= 4 and line[:3].isdigit():
                    code = line[:3]
                    if line[3:4] == b' ':  # final line
                        return b'\n'.join(lines) + b'\n'
        except BlockingIOError:
            time.sleep(0.01)
            continue

    return b'\n'.join(lines) + b'\n'


def conversation(s, script, read_fn = None):
    verbose = False
    for ph in script:
        if ph.say is not None:
            if verbose:
                print(">", repr(ph.say)) # pragma: no cover
            s.sendall(ph.say.encode())
        if read_fn:
            reply = read_fn(s, timeout=5).decode('utf8')
        else:
            reply = recv_until_newline(s, timeout=5).decode('utf8')

        if verbose:
            print("<", repr(reply))  # pragma: no cover
            print("wait:", repr(ph.wait)) # pragma: no cover
        if ph.wait is not None and ph.wait not in reply: # pragma: no cover
            raise ServerError('Not found {!r} in server reply {!r} to {!r}'.format(ph.wait, reply, ph.say)) # pragma: no cover
        if ph.expect is not None and ph.expect not in reply: # pragma: no cover
            raise ServerError('Not found {!r} in server reply {!r} to {!r}'.format(ph.expect, reply, ph.say)) # pragma: no cover
        if verbose:
            print("got it") # pragma: no cover

def starttls_imap(s):
    script = (
        phrase(None, '\n', None),
        phrase('a1 CAPABILITY\r\n', '\n', 'STARTTLS'),
        phrase('a2 STARTTLS\r\n','\n', None)
    )
    conversation(s, script)

def starttls_smtp(s):
    script = (
        phrase(None, '\n', None),
        phrase('EHLO www-security.com\n', '\n', 'STARTTLS'),
        phrase('STARTTLS\n','\n', None)
    )
    conversation(s, script, read_fn=recv_smtp)

def starttls_pop3(s):
    script = (
        phrase(None, '\n', None),
        phrase('STLS\n', '\n', None),
    )
    conversation(s, script)


def start_tls(s, method, port):

    port2method_map = {
        25: 'smtp',
        587: 'smtp',
        110: 'pop3',
        143: 'imap'
    }

    method_map ={
        'imap': starttls_imap,
        'smtp': starttls_smtp,
        'pop3': starttls_pop3
    }

    if method == 'no':
        return

    if method == 'auto':
        try:
            method = port2method_map[port]
        except KeyError:
            # no special handling needed
            return
    else:
        if method not in method_map:
            raise ValueError('Unknown starttls method {!r}'.format(method))

    return method_map[method](s)


def connect46(host, port, limit=5, ipv4=False):
    s = None
    family = socket.AF_INET if ipv4 else socket.AF_UNSPEC

    for res in socket.getaddrinfo(host, port, family, socket.SOCK_STREAM):
        af, socktype, proto, canonname, sa = res
        try:
            s = socket.socket(af, socktype, proto)
            s.settimeout(limit)
            s.connect(sa)
            return s
        except Exception as e:
            if s is not None:
                s.close()
            raise
    

def get_certificate_chain(host, name=None, port=443, insecure=False, starttls='auto', 
                          trusted_ca=None, limit=3, ipv4=False):
    

    name = name or host
    context = SSL.Context(method=SSL.TLS_CLIENT_METHOD)

    trusted_ca = trusted_ca or certifi.where()

    if insecure:
        context.set_verify(SSL.VERIFY_NONE)
    else:
        context.set_verify(SSL.VERIFY_PEER)
        context.load_verify_locations(trusted_ca)


    # s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    # s.settimeout(limit)
    try:
        # s.connect((host, port))
        s = connect46(host, port, limit=limit, ipv4=ipv4)
    except Exception as e:
        print(type(e))
        print(e)
        raise

    peername = s.getpeername()
    sock_host = peername[0]
    sock_port = peername[1]

    start_tls(s, starttls, port)
    conn = SSL.Connection(
        context, socket=s
    )
        
    # conn.settimeout(5)
    # conn.connect((host, port))

    conn.setblocking(1)    

    conn.set_tlsext_host_name(name.encode())
    
    conn.set_connect_state()

    try:
        conn.do_handshake()
    except SSL.Error as e:
        # rare case, e.g. RabbitMQ on 5671 which reset connection if client certificate is not sent
        # never happens on webservers
        if insecure and 'ssl/tls alert handshake failure' in str(e):
            print("# Server likely requires a client certificate (handshake failure)")
        else:
            raise
                
    if conn.get_client_ca_list():
        print("# Remote asks for a client certificate")

    chain = conn.get_peer_cert_chain()
    return sock_host, chain


def get_remote_certs(remote, name=None, insecure=False, starttls='auto', trusted_ca = None, limit = None, ipv4=False):
    # parse CERT address

    services = {
        'http': 80,
        'https': 443,
        'pop3': 110,
        'pop3s': 995,
        'imap': 143,
        'imap2': 143,
        'imaps': 993,
        'smtp': 25,
        'submissions': 465
    }

    if ':' in remote:
        try:
            (host, port) = remote.split(':')
        except ValueError:
            raise InvalidAddress('Invalid remote address. Valid examples: github.com or smtp.google.com:25')
    else:
        host = remote
        port = 443
    
    name = name or host

    try:
        port = services[port]
    except KeyError:
        try:
            port = int(port)
        except ValueError:
            raise InvalidAddress('Invalid remote address. Valid examples: github.com or smtp.google.com:25')

    certlist = get_certificate_chain(host, name=name, port=port, insecure=insecure, 
                                     starttls=starttls, trusted_ca=trusted_ca, limit=limit, ipv4=ipv4)
    
    # cert = load_certificate(FILETYPE_PEM, certificate)
    return certlist



from unittest import mock
from showcert.cli.showcert_main import main


snakeoil = '/etc/ssl/certs/ssl-cert-snakeoil.pem'

def test_main():
    with mock.patch('sys.argv', ['showcert_main.py', 'github.com']):
                    code = main()
                    assert code == 0
                    
def test_main_le():
    with mock.patch('sys.argv', ['showcert_main.py', ':le']):
                    code = main()
                    assert code == 0

def test_main_snakeoil():
    with mock.patch('sys.argv', ['showcert_main.py', snakeoil]):
                    code = main()
                    assert code == 1

def test_main_notacert():
    
    with mock.patch('sys.argv', ['showcert_main.py', "/etc/issue"]):
                    code = main()
                    assert code == 1



from pathlib import Path

from OpenSSL.crypto import load_certificate, FILETYPE_PEM
from OpenSSL.crypto import X509Store, X509StoreContext
import certifi
import pem

from .printcert import get_names_openssl
from .exceptions import InvalidCertificate


def wildcard_match(host: str, pattern: str) -> bool:
    if pattern.startswith('*.') and '.' in host:
        #wildcard comparison
        host_right = host.split('.', 1)[1]
        pattern_right = pattern.split('.',1)[1]
        return host_right == pattern_right        
    else:        
        return host == pattern

def verify_chain(chain, hostname=None, trusted_ca = None):

    trusted_ca = trusted_ca or certifi.where()

    # verify
    store = X509Store()

    raw_ca = Path(trusted_ca).read_bytes().rstrip()

    for _ca in pem.parse(raw_ca):
        if isinstance(_ca, pem._object_types.Certificate):
            store.add_cert(load_certificate(FILETYPE_PEM, str(_ca)))

    # verify and add each intermediate cert
    for _i in reversed(chain[1:]):
        
        _sctx = X509StoreContext(store, _i)
        try:
            _sctx.verify_certificate()
            store.add_cert(_i)
        except Exception as e:
            pass

    
    store_ctx = X509StoreContext(store, chain[0])
    store_ctx.verify_certificate()

    if hostname:

        names = get_names_openssl(chain[0])
        for _n in names:
            if wildcard_match(hostname, _n):
                return

        # not found
        raise InvalidCertificate('{} not found in {}'.format(hostname, ' '.join(names)))




#!/usr/bin/env python3

from cmath import phase
import os
import sys
import argparse
import re
import hashlib
import ssl
import socket
# import OpenSSL
import certifi
import glob
from datetime import datetime
from cryptography import x509

from urllib.parse import urlparse

from ..__about__ import __version__
from ..getremote import get_remote_certs
from ..exceptions import CertException, InvalidCertificate
from ..verifychain import verify_chain
from ..printcert import print_cert, print_full_cert, print_dnames, print_names
from ..processcert import process_cert

#import importlib.metadata
#version = importlib.metadata.version("showcert")

# do not  use CLI args globally
# args = None

def fix_cert(url: str) -> str:
    """ if https url: translate it to host """
    parsed = urlparse(url)    
    return parsed.hostname or url

def get_args():
    def_ca = certifi.where()
    def_limit = 5

    epilog = """Examples:  
  # just check remote certificate
  {me} example.com

  # check SMTP server certificate (autodetected: --starttls smtp )
  {me} smtp.google.com:25

  # save fullchain from google SMTP to local PEM file
  {me} --chain -o pem google.com > google-fullchain.pem
  
  # look for expiring letsencrypt certificates 
  # :le is alias for /etc/letsencrypt/live/*/fullchain.pem 
  {me} :le -q -w 20 || echo "expiring soon!"
  
    """.format(me='showcert')


    parser = argparse.ArgumentParser(description='Show local/remote SSL certificate info ver {version}'.format(version=__version__),
    formatter_class=argparse.RawTextHelpFormatter, epilog=epilog)
    parser.add_argument('CERT', nargs='+', help='path, - (stdin), ":le" (letsencrypt cert path), hostname or hostname:port')
    parser.add_argument('-i', '--insecure', default=False, action='store_true', help='Do not verify remote certificate')
    parser.add_argument('--output', '-o', choices=['brief', 'ext', 'full', 'names', 'dnames', 'pem', 'no'], default='brief', help='output format')
    parser.add_argument('-c','--chain', default=False, action='store_true', help='Show chain (not only server certificate)')
    parser.add_argument('-w', '--warn', default=None, metavar='DAYS', nargs='?', type=int, const=20, help='Warn about expiring certificates (def: 20 days)')
    parser.add_argument('-p', '--password', default=None, metavar='PASSWORD', help='Password (for PKCS#12 certificates)')

    g = parser.add_argument_group('Rarely needed options')
    g.add_argument('-q', '--quiet', default=False, action='store_true', help='Quiet mode, same as --output no')
    g.add_argument('-n', '--name', help='name for SNI (if not same as CERT host)')
    g.add_argument('-t', '--starttls', default='auto', metavar='METHOD', help='starttls method: auto (default, and OK almost always), no, imap, smtp, pop3')
    g.add_argument('-l', '--limit', default=def_limit, type=int, metavar='TIME', help='socket timeout (def: {})'.format(def_limit))
    g.add_argument('--ca', default=def_ca, help="path to trusted CA certificates, def: {}".format(def_ca))
    g.add_argument('--net', default=False, action='store_true',
                   help="Force network check (if you want to check host and have file/dir with same name in current directory)")
    g.add_argument('-4', default=False, dest='ipv4', action='store_true', help="use IPv4 only")


    return parser.parse_args()

def main():
    
    # global args 
    args = get_args()

    output = args.output if not args.quiet else 'no'


    if ':le' in args.CERT:
        args.CERT = glob.glob('/etc/letsencrypt/live/*/fullchain.pem')
    
    maxrc = 0
    for cert in args.CERT:
        try:
            rc = process_cert(
                CERT=fix_cert(cert), 
                name=args.name, 
                insecure=args.insecure, 
                warn=args.warn, 
                starttls=args.starttls,
                output=output,
                force_network=args.net,
                trusted_ca=args.ca,
                chain=args.chain,
                limit=args.limit,
                password=args.password,
                ipv4=args.ipv4)
            maxrc = max(maxrc, rc)
        except (CertException, ValueError, OSError) as e:
            print("{}: {}".format(cert, e))
            maxrc=1
    return(maxrc)


if __name__ == '__main__':
    sys.exit(main())
    



#!/usr/bin/env python

import argparse
import ipaddress
import datetime
import os
import sys
import ipaddress

from typing import List

from cryptography import x509
from cryptography.x509 import Certificate
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa




from .. import __version__


# Gists:
# Create a self-signed x509 certificate with python cryptography library
#     https://gist.github.com/bloodearnest/9017111a313777b9cce5
# Making a certificate authority (CA) with python cryptography
#     https://gist.github.com/major/8ac9f98ae8b07f46b208
#
# https://github.com/ikreymer/certauth/blob/master/certauth/certauth.py


def generate_cert(hostnames: list, ip_addresses: list = None, 
                    cakey=None, cacert=None, days=365, bits=2048, 
                    ca=False) -> tuple:
    """Generates self signed certificate for a hostname, and optional IP addresses."""
    
    # Generate our key
    privkey = rsa.generate_private_key(
        public_exponent=65537,
        key_size=bits,
        backend=default_backend(),
    )

    name = x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, hostnames[0])
    ])

    # best practice seem to be to include the hostname in the SAN, which *SHOULD* mean COMMON_NAME is ignored.    
    alt_names = [x509.DNSName(h) for h in hostnames]
    
    # allow addressing by IP, for when you don't have real DNS (common in most testing scenarios 
    if ip_addresses:
        for addr in ip_addresses:
            # openssl wants DNSnames for ips...
            # we add above: alt_names.append(x509.DNSName(addr))
            # ... whereas golang's crypto/tls is stricter, and needs IPAddresses
            # note: older versions of cryptography do not understand ip_address objects
            alt_names.append(x509.IPAddress(ipaddress.ip_address(addr)))
    
    san = x509.SubjectAlternativeName(alt_names)
    
    # path_len=0 means this cert can only sign itself, not other certs.
    
    now = datetime.datetime.now(datetime.timezone.utc)





    builder = x509.CertificateBuilder() \
        .subject_name(name) \
        .public_key(privkey.public_key()) \
        .serial_number(x509.random_serial_number()) \
        .not_valid_before(now) \
        .not_valid_after(now + datetime.timedelta(days=days)) \

    # Add Subject Key Identified (SKI)
    ski = x509.SubjectKeyIdentifier.from_public_key(privkey.public_key())
    builder = builder.add_extension(ski, critical=False)

    if ca:
        print("Generate CA certificate")
        basic_constraints = x509.BasicConstraints(ca=True, path_length=None)
        builder = builder.add_extension(basic_constraints, True)

        key_usage = x509.KeyUsage(
            digital_signature=False,
            content_commitment=False,
            key_encipherment=False,
            data_encipherment=False,
            key_agreement=False,
            key_cert_sign=True,
            crl_sign=True,
            encipher_only=False,
            decipher_only=False,
        )
        builder = builder.add_extension(key_usage, critical=True)

#                    crypto.X509Extension(b"basicConstraints",
#                                 True,
#                                b"CA:TRUE, pathlen:0"),

    else:
        basic_constraints = x509.BasicConstraints(ca=False, path_length=None)
        builder = builder.add_extension(basic_constraints, False) \
            .add_extension(san, False)

    # Issuer
    if cacert and cakey:
        builder = builder.issuer_name(cacert.issuer)
    else:
        builder = builder.issuer_name(name)

    # SIGN
    if cakey:
        aki = x509.AuthorityKeyIdentifier.from_issuer_public_key(cakey.public_key())
        builder = builder.add_extension(aki, critical=False)

        # sign with CA private key
        cert = builder.sign(private_key=cakey, 
            algorithm=hashes.SHA256(), 
            backend=default_backend())
    else:       
        aki = x509.AuthorityKeyIdentifier.from_issuer_public_key(privkey.public_key())
        builder = builder.add_extension(aki, critical=False)

        # self-sign
        cert = builder.sign(private_key=privkey, 
            algorithm=hashes.SHA256(), 
            backend=default_backend())
    

    # cert_pem = cert.public_bytes(encoding=serialization.Encoding.PEM)
    # key_pem = privkey.private_bytes(
    #    encoding=serialization.Encoding.PEM,
    #    format=serialization.PrivateFormat.TraditionalOpenSSL,
    #    encryption_algorithm=serialization.NoEncryption(),
    #)
    # return cert_pem, key_pem
    return cert, privkey

def get_args():

    def_days = 365
    def_bits = 2048

    epilog = '''    
Examples:\n

# make simple self-signed cert and key in one file example.com.pem  
gencert example.com www.example.com 


# Your own CA:
# make CA cert and key
gencert --ca  --cert ca.pem --key ca-priv.pem "My CA"

# make host cert and sign CA cert
gencert --cacert ca.pem --cakey ca-priv.pem example.com www.example.com
'''

    parser = argparse.ArgumentParser(
        description=f"gencert version {__version__}",
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--cert', help='save new certificate to this file')
    parser.add_argument('--key', help='save private key to this file (optional)')
    parser.add_argument('hostnames', metavar='HOST', nargs='+', help='Hostnames and/or IP addresses')

    g = parser.add_argument_group('CA operations')
    g.add_argument('--ca', default=False, action='store_true', help='Generate --cert/--key for CA, not usual cert')
    g.add_argument('--cakey', metavar='ca.crt', help='read CA key to sign new cert (optional)')
    g.add_argument('--cacert', metavar='file.pem', help='read CA cert to sign new cert (optional)')

    g = parser.add_argument_group('Options')
    g.add_argument('-d', '--days', type=int, metavar='DAYS', default=def_days,
                   help=f'expire in DAYS days ({def_days})')
    g.add_argument('-b', '--bits', type=int, metavar='BITS', default=def_bits,
                   help=f'Key size in bits ({def_bits})')


    parser.epilog = epilog

    return parser.parse_args()

def good_filename(path):
    return path.replace(' ','-')


def change_extension(filename, new_extension):
    base_name, _ = os.path.splitext(filename)
    return base_name + new_extension

def load_privkey(files: List[str]):
    for file in files:
        if file is None:
            continue
        with open(file, 'rb') as fh:
            try:
                privkey = serialization.load_pem_private_key(fh.read(), password=None)
                return privkey
            except ValueError:
                pass


def save_key(fh, key: rsa.RSAPrivateKey):
    fh.write(key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption()))

def main() -> int:
    args = get_args()
    ca_privkey = None
    ca_cert = None
    ipaddresses = list()

    certfile = args.cert
    keyfile = args.key

    if certfile is None:
        certfile = good_filename(args.hostnames[0] + '.pem')
    
    if not keyfile:
        if args.ca:
            # we generate CA certs, key is in different file
            keyfile = change_extension(certfile, '.key')
        else:
            keyfile = certfile

    if args.cacert:
        with open(args.cacert, 'rb') as fh:
            ca_cert = x509.load_pem_x509_certificate(fh.read(), default_backend())


    if args.cacert:
        ca_privkey = load_privkey([args.cakey, args.cacert, change_extension(args.cacert, '.key')])

    #if args.cakey:
    #    with open(args.cakey, 'rb') as fh:
            # ca_privkey = rsa.PrivateKey.load_pkcs1(fh.read())
    #        ca_privkey = serialization.load_pem_private_key(fh.read(), password=None)


    for h in args.hostnames:
        try:
            ipaddress.ip_address(h)
            ipaddresses.append(h)
        except ValueError:
            pass


    cert, key = generate_cert(hostnames = args.hostnames, 
                    ip_addresses = ipaddresses,
                    days=args.days, bits=args.bits,
                    cakey=ca_privkey, cacert=ca_cert,
                    ca=args.ca)


    with open(certfile, "wb") as fh:
        fh.write(cert.public_bytes(encoding=serialization.Encoding.PEM))
        if keyfile == certfile:
            save_key(fh, key)
    
    if keyfile != certfile:
        # different key file
        with open(keyfile, "wb") as fh:
            save_key(fh, key)

    return 0

if __name__ == '__main__':
    sys.exit(main())



from showcert import process_cert
from OpenSSL import SSL

import pytest

class TestClient:
    def test_client(self):
        srv = "client.badssl.com"
        process_cert(srv)

    def test_handshake_failure(self, monkeypatch):
        def fake_handshake(self):
            original_handshake(self)
            raise SSL.Error([('SSL routines', '', 'ssl/tls alert handshake failure')])

        original_handshake = SSL.Connection.do_handshake
        monkeypatch.setattr("OpenSSL.SSL.Connection.do_handshake", fake_handshake)
        process_cert("client.badssl.com", insecure=True)




from showcert import process_cert
from unittest import mock
import io

class TestShowcertLocal():

    showcert = 'showcert'
    gencert = 'gencert'
    snakeoil = '/etc/ssl/certs/ssl-cert-snakeoil.pem'
    ca_certs = [ '/etc/ssl/certs/DigiCert_Global_Root_CA.pem', '/etc/ssl/certs/Amazon_Root_CA_1.pem', '/etc/ssl/certs/Amazon_Root_CA_2.pem']

    def test_snakeoil(self):
        rc = process_cert(CERT=self.snakeoil)
        assert(rc == 1)

        rc = process_cert(CERT=self.snakeoil, insecure=True)
        assert(rc == 0)

    def test_snakeoil_output(self):
        rc = process_cert(CERT=self.snakeoil, output='no', insecure=True)
        assert(rc == 0)

        rc = process_cert(CERT=self.snakeoil, output='nosuchformat', insecure=True)
        assert(rc == 0)


    def test_ca(self):
        for ca in self.ca_certs:
            print("test:", ca)
            rc = process_cert(CERT=ca)
            assert(rc == 0)

    def test_print(self):
        rc = process_cert(CERT=self.ca_certs[0], output='full')
        assert(rc == 0)
        rc = process_cert(CERT=self.ca_certs[0], output='ext')
        assert(rc == 0)
        rc = process_cert(CERT=self.snakeoil, output='names', insecure=True)
        assert(rc == 0)
        rc = process_cert(CERT=self.snakeoil, output='dnames', insecure=True)
        assert(rc == 0)
        rc = process_cert(CERT=self.snakeoil, output='pem', insecure=True)
        assert(rc == 0)
        rc = process_cert(CERT=self.snakeoil, output='no', insecure=True)
        assert(rc == 0)

    def test_stdin(self, monkeypatch):
        with open(self.ca_certs[0], "rb") as f:  # Read the certificate file
            mock_input = f.read()

        mock_stdin = mock.Mock()
        mock_stdin.buffer = io.BytesIO(mock_input)
        monkeypatch.setattr("sys.stdin", mock_stdin)

        rc = process_cert(CERT='-')
        assert rc == 0



from unittest import mock
from showcert.cli.gencert_main import main
import os
def test_cacert():
    with mock.patch('sys.argv', ['gencert_main.py', 
        '--ca', '--cert', '/tmp/ca.pem', '--key', '/tmp/ca-priv.pem', "My CA"]):                    
                    code = main()
                    assert code == 0


def test_cacert_combined():
    with mock.patch('sys.argv', ['gencert_main.py', 
        '--ca', '--cert', '/tmp/ca2.pem', "My CA"]):                    
                    code = main()
                    assert code == 0


def test_cert():
    with mock.patch('sys.argv', ['gencert_main.py', 
        '--cacert', '/tmp/ca.pem', '--cakey', '/tmp/ca-priv.pem', 'example.com', 'www.example.com', '0.0.0.1']):
                    code = main()
                    assert code == 0
                    os.unlink('example.com.pem')

def test_cert_combined():
    with mock.patch('sys.argv', ['gencert_main.py', 
        '--cacert', '/tmp/ca2.pem', 'example.com', 'www.example.com']):
                    code = main()
                    assert code == 0
                    os.unlink('example.com.pem')



from cryptography.hazmat.primitives.serialization import pkcs12
from cryptography import x509
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID

from showcert import process_cert
import datetime
from unittest import mock
import os

class TestShowcertPKCS12():

    showcert = 'showcert'
    gencert = 'gencert'
    testcert_p12 = '/tmp/testcert.p12'
    testchain_p12 = '/tmp/testchain.p12'    
    password = 'mysecret'


    def make_p12chain(self):
        # Generate root CA
        root_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
        root_name = x509.Name([
            x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Root CA"),
            x509.NameAttribute(NameOID.COMMON_NAME, "Root CA"),
        ])
        root_cert = (
            x509.CertificateBuilder()
            .subject_name(root_name)
            .issuer_name(root_name)
            .public_key(root_key.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(datetime.datetime.utcnow())
            .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=3650))
            .add_extension(x509.BasicConstraints(ca=True, path_length=None), critical=True)
            .sign(root_key, hashes.SHA256())
        )

        # Generate intermediate CA
        int_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
        int_name = x509.Name([
            x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Intermediate CA"),
            x509.NameAttribute(NameOID.COMMON_NAME, "Intermediate CA"),
        ])
        int_cert = (
            x509.CertificateBuilder()
            .subject_name(int_name)
            .issuer_name(root_name)
            .public_key(int_key.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(datetime.datetime.utcnow())
            .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=730))
            .add_extension(x509.BasicConstraints(ca=True, path_length=0), critical=True)
            .sign(root_key, hashes.SHA256())
        )

        # Generate end-entity certificate
        ee_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
        ee_name = x509.Name([
            x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "My Company"),
            x509.NameAttribute(NameOID.COMMON_NAME, "mysite.com"),
        ])
        ee_cert = (
            x509.CertificateBuilder()
            .subject_name(ee_name)
            .issuer_name(int_name)
            .public_key(ee_key.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(datetime.datetime.utcnow())
            .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=365))
            .add_extension(x509.BasicConstraints(ca=False, path_length=None), critical=True)
            .sign(int_key, hashes.SHA256())
        )

        # Create PKCS12 with chain (end-entity cert + intermediate + root)
        p12_data = pkcs12.serialize_key_and_certificates(
            name=b"cert-with-chain",
            key=ee_key,
            cert=ee_cert,
            cas=[int_cert, root_cert],
            encryption_algorithm=serialization.BestAvailableEncryption(self.password.encode())
        )

        with open(self.testchain_p12, "wb") as f:
            f.write(p12_data)

        print(f"Generated certificate with chain: {self.testchain_p12}")
        

    def make_p12(self):
        # Generate private key
        private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

        # Create self-signed cert
        subject = issuer = x509.Name([
            x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "California"),
            x509.NameAttribute(NameOID.LOCALITY_NAME, "San Francisco"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "My Company"),
            x509.NameAttribute(NameOID.COMMON_NAME, "mysite.com"),
        ])

        cert = x509.CertificateBuilder().subject_name(
            subject
        ).issuer_name(
            issuer
        ).public_key(
            private_key.public_key()
        ).serial_number(
            x509.random_serial_number()
        ).not_valid_before(
            datetime.datetime.utcnow()
        ).not_valid_after(
            datetime.datetime.utcnow() + datetime.timedelta(days=365)
        ).add_extension(
            x509.BasicConstraints(ca=True, path_length=None), critical=True,
        ).sign(private_key, hashes.SHA256())

        # Export as PKCS12 (.p12)
        password = self.password.encode()  # Change this!
        p12 = serialization.pkcs12.serialize_key_and_certificates(
            name=b"self-signed-cert",
            key=private_key,
            cert=cert,
            cas=None,
            encryption_algorithm=serialization.BestAvailableEncryption(self.password.encode())
        )

        with open(self.testcert_p12, "wb") as f:
            f.write(p12)

        print("Generated self_signed.p12")


    def test_p12(self):

        self.make_p12()

        rc = process_cert(CERT=self.testcert_p12, password=self.password)
        assert(rc == 1)

        rc = process_cert(CERT=self.testcert_p12, password=self.password, insecure=True)
        assert(rc == 0)

        # os.unlink(self.testcert_p12)

    def test_chain_p12(self):

        self.make_p12chain()

        rc = process_cert(CERT=self.testchain_p12, password=self.password, insecure=True)
        assert(rc == 0)

        # os.unlink(self.testcert_p12)



import time
from showcert import process_cert
from showcert.exceptions import InvalidAddress
import random
import string
import pytest

class TestShowcertRemote():

    showcert = 'showcert'

    sites = [ 'github.com', 'okerr.com', 'www-security.com', 'gmail.com' ]
    wildcard_sites = [ 'www.badssl.com' ]
    badssl_sites = ['expired.badssl.com', 'wrong.host.badssl.com', 'self-signed.badssl.com', 'untrusted-root.badssl.com']
    pop3_sites = ['pop.yandex.ru:110']
    pop3s_sites = ['pop.yandex.ru:995', 'pop.gmail.com:995']
    imap_sites = ['imap.yandex.ru:143']
    imaps_sites = ['imap.yandex.ru:993', 'imap.gmail.com:993']
    smtp_sites = ['smtp.yandex.ru:25']
    ev_sites = ['www.bankofamerica.com', 'www.hsbc.com.hk']
    ip46_sites = ['google.com', 'microsoft.com']

    def test_https(self):
        for site in self.sites:
            print("test site {}".format(site))
            code = process_cert(CERT=site)
            assert code == 0

    def test_https_chain(self):
        # warn if expires "too soon" in 2000 days
        code = process_cert(CERT=self.sites[0], chain=True)
        assert code == 0
        code = process_cert(CERT=self.sites[0], chain=True, output='full')
        assert code == 0
        code = process_cert(CERT=self.sites[0], chain=True, output='pem')
        assert code == 0
        code = process_cert(CERT=self.sites[0], chain=True, output='names')
        assert code == 0
        code = process_cert(CERT=self.sites[0], chain=True, output='dnames')
        assert code == 0


    def test_https_methods(self):
        code = process_cert(CERT=self.sites[0], starttls='no')
        assert code == 0

        with pytest.raises(ValueError):
            process_cert(CERT=self.sites[0], starttls='nosuchmethod')

    def test_invalid_address(self):
        with pytest.raises(InvalidAddress):
            process_cert("aa:bb.com:443")
        with pytest.raises(InvalidAddress):
            process_cert("github.com:notaport")

    def test_https_warn(self):
        # warn if expires "too soon" in 2000 days
        code = process_cert(CERT=self.sites[0], warn=2000, force_network=True)
        assert code == 2

    def test_wildcard(self):
        for site in self.wildcard_sites:
            code = process_cert(CERT=site)
            assert code == 0

    def test_badssl(self):
        for site in self.badssl_sites:
            code = process_cert(CERT=site)
            assert code == 1

    def test_badssl_ignore(self):
        for site in self.badssl_sites:
            code = process_cert(CERT=site, insecure=True)
            assert code == 0

    def test_ev(self):
        for site in self.ev_sites:
            code = process_cert(CERT=site, insecure=True)
            assert code == 0

    def test_smtp(self):
        for site in self.smtp_sites:
            code = process_cert(CERT=site)
            assert code == 0

    def test_pop3(self):
        for site in self.pop3_sites:
            code = process_cert(CERT=site)
            assert code == 0

    def test_pop3s(self):
        for site in self.pop3s_sites:
            code = process_cert(CERT=site)
            assert code == 0

    def test_imap(self):
        for site in self.imap_sites:
            code = process_cert(CERT=site)
            assert code == 0

    def test_imaps(self):
        for site in self.imaps_sites:
            code = process_cert(CERT=site)
            assert code == 0

    def test_ip46(self):
        for site in self.ip46_sites:
            code = process_cert(CERT=site, ipv4=False)
            assert code == 0

    def test_timeout(self):
        test_start = time.time()
        code = process_cert(CERT='0.0.0.1', limit=2)
        test_end = time.time()
        assert code == 1
        assert test_end - test_start >= 1

    def test_nosuchdomain(self):
        name = 'nosuchdomain-' + ''.join(random.choices(string.ascii_lowercase, k=20)) + '.com'
        code = process_cert(CERT=name)
        assert code == 1

    def test_connrefused(self):
        code = process_cert(CERT='localhost:1234')
        assert code == 1


from OpenSSL.crypto import X509
from cryptography import x509
from OpenSSL.crypto import dump_certificate, FILETYPE_PEM
from cryptography.hazmat.backends import default_backend


def convert_openssl_to_cryptography(openssl_cert: X509) -> x509.Certificate:
    # Convert OpenSSL.crypto.X509 to PEM format
    pem_data = dump_certificate(FILETYPE_PEM, openssl_cert)

    # Load PEM data into cryptography.x509.Certificate
    certificate = x509.load_pem_x509_certificate(pem_data, default_backend())

    return certificate



from showcert.cli.gencert_main import generate_cert, save_key
from showcert import process_cert
from cryptography.hazmat.primitives import serialization
from pathlib import Path
import os

class TestGencert():
    path: Path

    def setup_class(self):
        tmp_path = Path(f"/tmp/ca-{os.getpid()}")
        tmp_path.mkdir()

        # print("method:", method)
        
        self.path = tmp_path

        self.ca_cert_path = self.path / 'ca.pem'
        self.ca_key_path = self.path / 'ca-priv.pem'
        self.cert_path = self.path / 'example.pem'
        self.key_path = self.path / 'example-priv.pem'
        self.badchain_path = self.path / 'badchain.pem'

        
        cacert, cakey = generate_cert(["Test CA"], ca=True, days=1)
        with open(self.ca_cert_path, "wb") as fh:
            fh.write(cacert.public_bytes(encoding=serialization.Encoding.PEM))        
        with open(self.ca_key_path, "wb") as fh:
            save_key(fh, cakey)

        self.cacert = cacert
        self.cakey = cakey

        cert, key = generate_cert(["example.com", "www.example.com"], cakey=self.cakey, cacert=self.cacert)
        with open(self.cert_path, "wb") as fh:
            fh.write(cert.public_bytes(encoding=serialization.Encoding.PEM))        
        with open(self.key_path, "wb") as fh:
            save_key(fh, key)

        badchain, _ = generate_cert(["example.com", "www.example.com"], cakey=self.cakey, cacert=self.cacert)
        with open(self.badchain_path, "wb") as fh:
            fh.write(badchain.public_bytes(encoding=serialization.Encoding.PEM))
            fh.write(cert.public_bytes(encoding=serialization.Encoding.PEM))        


    def test_certs(self):

        rc = process_cert(CERT=self.cert_path)
        assert(rc == 1)

        rc = process_cert(CERT=self.cert_path, insecure=True)
        assert(rc == 0)

        rc = process_cert(CERT=self.cert_path, trusted_ca=self.ca_cert_path)
        assert(rc == 0)

    def test_badchain(self):
        rc = process_cert(CERT=self.badchain_path)
        assert(rc == 1)        

    def teardown_class(self):        
        self.ca_cert_path.unlink()
        self.ca_key_path.unlink()
        self.cert_path.unlink()
        self.key_path.unlink()
        self.badchain_path.unlink()
        self.path.rmdir()




from datetime import datetime
from cryptography import x509
from cryptography.x509.extensions import ExtensionNotFound

from OpenSSL.crypto import X509, dump_certificate, FILETYPE_TEXT
from typing import List, Optional


from .utils import convert_openssl_to_cryptography
from .exceptions import InvalidCertificate


def get_extension_value(cert: x509.Certificate, oid):
    try:
        ext = cert.extensions.get_extension_for_oid(oid)
        return ext.value
    except x509.ExtensionNotFound:
        return None


def get_names_openssl(crt: X509):
    crt_crypto = convert_openssl_to_cryptography(crt)
    return get_names(crt_crypto)



def get_cn(crt: x509.Certificate):
    for attribute in crt.subject:
        if attribute.oid == x509.NameOID.COMMON_NAME:  # Check if the attribute is CN
            return attribute.value

def get_names(crt: x509.Certificate) -> List[str]:
    """ return CN subject of certificate + all DNS names in SAN extension """

    names = list()
    cn = get_cn(crt)
    if cn:
        names.append(cn)

    SAN = get_extension_value(crt, x509.OID_SUBJECT_ALTERNATIVE_NAME)
    if SAN is None:
        return names
    names.extend([name.value for name in SAN if isinstance(name, x509.DNSName)])
    return names

def is_self_signed(crt: x509.Certificate):
    return crt.issuer == crt.subject

def is_CA(crt: x509.Certificate):
    # Get the extensions from the certificate
    extensions = crt.extensions

    # Look for the BasicConstraints extension
    basic_constraints = next((ext.value for ext in extensions if isinstance(ext.value, x509.BasicConstraints)), None)

    # If BasicConstraints is not present, or if it is present and CA is set to True, it is a CA certificate
    return basic_constraints is None or basic_constraints.ca

def is_EV(crt: x509.Certificate) -> bool:
    EV_OIDS = ['2.23.140.1.1']
    try:
        policies = crt.extensions.get_extension_for_class(x509.CertificatePolicies).value
        for policy in policies:
            if policy.policy_identifier.dotted_string in EV_OIDS:
                return True  # Found a policy OID that matches a known EV OID

    except x509.ExtensionNotFound:
        # No certificate policies extension found
        pass
    
    return False


def print_full_cert(crt):
    print(dump_certificate(FILETYPE_TEXT, crt).decode())

def print_names(crt: X509):
    # expects openssl crt!
    cc = convert_openssl_to_cryptography(crt)
    names = get_names(cc)

    print(' '.join(names))

def print_dnames(crt):
    cc = convert_openssl_to_cryptography(crt)
    names = get_names(cc)
    print('-d', ' -d '.join(names))

def hexify(b: bytes) -> str:
    hex_string = b.hex()
    return ":".join(hex_string[i:i+2] for i in range(0, len(hex_string), 2)).upper()

def print_cert(crt: X509, fmt='brief', addr=None, path=None, verified=False):

    def tlist2str(tlist):
        return ' '.join([ '{}={}'.format(t[0].decode(), t[1].decode()) for t in tlist ])

    tags = list()

    crypto_crt = convert_openssl_to_cryptography(crt)

    if is_self_signed(crypto_crt):
        tags.append('[SELF-SIGNED]')

    if is_CA(crypto_crt):
        tags.append('[CA]')

    if is_EV(crypto_crt):
        tags.append('[EV]')

    if verified:
        tags.append('[CHAIN-VERIFIED]')


    nbefore = datetime.strptime(crt.get_notBefore().decode(), '%Y%m%d%H%M%SZ')
    nafter = datetime.strptime(crt.get_notAfter().decode(), '%Y%m%d%H%M%SZ')
    daysold = (datetime.now() - nbefore).days
    daysleft = (nafter - datetime.now()).days
    issuer = tlist2str(crt.get_issuer().get_components())
    fingerprint = crt.digest("sha256").decode("utf-8")

    #ski = crt.extensions.get_extension_for_oid(x509.ExtensionOID.SUBJECT_KEY_IDENTIFIER) \
    #    .value.digest.hex()


    subject_key_id = get_extension_value(crypto_crt, x509.OID_SUBJECT_KEY_IDENTIFIER)
    authority_key_id = get_extension_value(crypto_crt, x509.OID_AUTHORITY_KEY_IDENTIFIER)

    names = get_names(crypto_crt)

    if addr:
        print("IP:", addr)
    if path:
        print("Path:", path)
    print("Names:", ' '.join(names))
    print("notBefore: {nbefore} ({days} days old)".format(nbefore=nbefore, days=daysold))
    print("notAfter: {nafter} ({days} days left)".format(nafter=nafter, days = daysleft))
    print("Issuer:", issuer)


    try:
        ku = crypto_crt.extensions.get_extension_for_class(x509.KeyUsage).value
        
        if ku.key_cert_sign:
            tags.append('[keyCertSign]')
        if ku.crl_sign:
            tags.append('[crlSign]')
        if ku.digital_signature:
            tags.append('[digitalSignature]')

    except ExtensionNotFound:
        pass

    if fmt.startswith('ext'):
        print("Fingerprint (sha256):", fingerprint)
        if subject_key_id:
            print("Subject KI:", hexify(subject_key_id.digest))
        if authority_key_id:
            print("Authority KI:", hexify(authority_key_id.key_identifier))

    if tags:
        print("Tags:", ' '.join(tags))
